from .__about__ import (
    __author__,
    __commit__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
    __version__,
)

from .stim_makers import RVvGVStimMaker, Two_v_Five_StimMaker
from . import main
